import base64
import csv
import io
import itertools
import logging
import os
from datetime import date, datetime
from uuid import UUID
from zipfile import ZipFile

from PySimpleGUI import POPUP_BUTTONS_OK_CANCEL, PySimpleGUI as sg, POPUP_BUTTONS_NO_BUTTONS
from satcfdi import csf
from satcfdi.accounting import EmailManager
from satcfdi.accounting.models import EstadoComprobante
from satcfdi.accounting.process import complement_invoices
from satcfdi.create.cfd import cfdi40
from satcfdi.create.cfd.catalogos import MetodoPago, TipoDeComprobante
from satcfdi.exceptions import ResponseError
from satcfdi.models import DatePeriod
from satcfdi.pacs import Accept
from satcfdi.pacs.sat import SAT, EstadoSolicitud
from xlsxwriter.exceptions import XlsxFileError

from . import __version__, TEMP_DIRECTORY, ARCHIVOS_DIRECTORY, DATA_DIRECTORY, METADATA_FILE
from .client_validation import validar_client, clientes_generar_txt
from .environments import facturacion_environment
from .file_data_managers import ClientsManager, FacturasManager
from .gui_functions import generate_ingresos, pago_factura, exportar_facturas, archivos_folder, period_desc, parse_fecha_pago, parse_importe_pago, preview_cfdis, center_location, \
    CALENDAR_FECHA_FMT, ConsoleErrors, \
    generate_ajustes, generar_depositos, calculate_declaracion_provisional, calculate_diot
from .layout import make_layout, ActionButtonManager, TipoRecuperar, SearchOptions
from .localdb import LocalDBSatCFDI, StatusState
from .log_tools import header_line, print_yaml, to_yaml
from .mycfdi import MyCFDI
from .utils import random_string, to_date_period, load_certificate, to_int, cert_info, add_month, to_uuid, open_file, OS

logging.getLogger("weasyprint").setLevel(logging.ERROR)
logging.getLogger("fontTools").setLevel(logging.ERROR)

logger = logging.getLogger(__name__)


def open_launch_window():
    layout = [[sg.Text("New Window", key="new")]]
    window = sg.Window("Launch Window", layout, modal=True, size=(300, 300))
    window.read(timeout=1000)

    return window


class FacturacionGUI:
    def __init__(self, config):
        os.makedirs(TEMP_DIRECTORY, exist_ok=True)
        self.email_manager = EmailManager(
            **config['email']
        )
        pac = config['pac']
        pac_module, pac_class = pac['type'].split(".")
        mod = __import__(f"satcfdi.pacs.{pac_module}", fromlist=[pac_class])
        self.pac_service = getattr(mod, pac_class)(
            **pac['args']
        )

        def load_emisor(data):
            return {
                "csd": load_certificate(data['csd']),
                "fiel": load_certificate(data['fiel']),
            }

        if emisores := config.get('emisores'):
            self.emisores = {
                rfc: load_emisor(csd) for rfc, csd in emisores.items()
            }
        else:
            self.emisores = None

        self.rfc_prediales = config['rfc_prediales']

        self._all_invoices = None
        self.local_db = LocalDBSatCFDI(
            base_path=DATA_DIRECTORY,
            enviar_a_partir=config['enviar_a_partir'],
            pagar_a_partir=config['pagar_a_partir']
        )

        MyCFDI.local_db = self.local_db
        MyCFDI.base_dir = ARCHIVOS_DIRECTORY

        self.window = sg.Window(
            f"Facturación Mensual CFDI 4.0",
            make_layout(list(self.emisores.keys()), self.local_db),
            size=(1280, 720),
            resizable=True,
            font=("Courier New", 10, "bold"),
            ttk_theme="default",
            margins=(0, 0),
            # use_custom_titlebar=True,
            titlebar_font=("Courier New", 11, "bold"),
            finalize=True,
            scaling=config.get('scaling', 1.0),
        )
        self.has_focus = True

        self.action_button_manager = ActionButtonManager(
            button=self.window["crear_facturas"],
            preview=self.window["ver_preview"],
        )
        self.console = self.window["console"]
        self.set_inputs()

        self.window.bind("<FocusIn>", "_focus_in")
        self.window.bind("<FocusOut>", "_focus_out")

        for t in ('serie', 'folio', 'serie_pago'):
            self.window[t].bind("<Return>", "_enter")
            self.window[t].bind("<FocusOut>", "_enter", propagate=False)

        for t in ('facturas_periodo', 'emitidas_search', 'recibidas_search', 'ajustes_periodo', 'periodo'):
            self.window[t].bind("<Return>", "_enter")

        modifier_key = "Command" if OS.get_os() == OS.MACOS else "Control"

        for t in ('facturas_table', 'clientes_table', 'emitidas_table', 'recibidas_table', 'correos_table', 'ajustes_table', 'depositos_table', 'solicitudes_table'):
            self.window[t].bind(f'<{modifier_key}-a>', '+select_all')
            self.window[t].bind('<BackSpace>', '+delete')  # BackSpace
            # self.window[t].bind('<Double-Button-1>', '_enter')
            self.window[t].bind('<Return>', '+enter')

    def run(self):
        self.main_loop()
        self.window.close()

    def initial_screen(self):
        self.header("ACERCA DE")

        def emisor_info(data):
            return {
                "csd": cert_info(data['csd']),
                "fiel": cert_info(data['fiel']),
            }

        print_yaml({
            "version": __version__.__version__,
            "facturacion": "CFDI 4.0",
            "pac_service": {
                "Type": type(self.pac_service).__name__,
                "Rfc": self.pac_service.RFC,
                "Environment": str(self.pac_service.environment)
            },
            "emisores": {rfc: emisor_info(data) for rfc, data in self.emisores.items()},
        })

    def get_all_invoices(self):
        if not self._all_invoices:
            self._all_invoices = MyCFDI.get_all_cfdi()
        return self._all_invoices

    def add_created_invoice(self, invoice: MyCFDI):
        self._all_invoices[invoice.uuid] = invoice
        complement_invoices(self._all_invoices, invoice)

    def generate_invoice(self, invoice):
        ref_id = random_string()
        title = 'Generando factura'

        # Add Serie and Folio and signature
        folio = None
        if 'Serie' not in invoice:
            invoice['Serie'] = self.local_db.serie()
            folio = self.local_db.folio()
            invoice['Folio'] = str(folio)

        cfdi40.Comprobante.sign(
            invoice,
            self.emisores[invoice['Emisor']['Rfc']]['csd']
        )

        attempts = 3
        try:
            for i in self.progress_iterate(
                    title, range(attempts), lambda r: f'Intentando de nuevo... Intento {r + 1} de {attempts}', skip_first=True, delay=2000
            ):
                try:
                    res = self.pac_service.stamp(
                        cfdi=invoice,
                        accept=Accept.XML_PDF,
                        ref_id=ref_id
                    )
                except Exception as ex:
                    self.show_console()
                    message = f"Error al generar factura: {invoice.get('Serie')}{invoice.get('Folio')} {invoice['Receptor']['Rfc']}"
                    message += f"\nIntento {i + 1} de {attempts}"

                    logger.exception(message)
                    print(message)
                    if isinstance(ex, ResponseError):
                        logger.error(f"Status Code: {ex.response.status_code}")
                        print(f"Status Code: {ex.response.status_code}")
                        logger.error(f"Response: {ex.response.text}")
                        print(f"Response: {ex.response.text}")
                    continue

                if folio is not None:
                    self.set_folio(folio + 1)
                cfdi = MyCFDI.move_to_folder(res.xml, pdf_data=res.pdf)
                self.add_created_invoice(cfdi)
                return cfdi
        finally:
            self.progress_cancel(title)

    def set_serie(self, serie: str = None):
        if serie:
            self.local_db.serie_set(serie.strip())
        self.set_inputs()

    def set_serie_pago(self, serie: str = None):
        self.local_db.serie_pago_set(serie.strip())
        self.set_inputs()

    def set_folio(self, folio: int = None):
        if folio:
            self.local_db.folio_set(folio)
        self.set_inputs()

    def set_inputs(self):
        folio = self.local_db.folio()
        serie = self.local_db.serie()
        serie_pago = self.local_db.serie_pago()

        self.window['serie'].update(serie)
        self.window['folio'].update(folio)
        self.window['serie_folio'].update(f"{serie}{folio}")

        self.window['serie_pago'].update(serie_pago)
        self.window['folio_pago'].update(
            "Se usara el mismo folio de la factura que se paga" if serie_pago else "Se usar serie y folio de proxima factura"
        )

    def nueva_solicitud(self, values):
        rfc = values["solicitudes_rfc"]

        sat_service = SAT(signer=self.emisores[rfc]['fiel'])
        tipo_recuperar = values["tipo_recuperar"]

        args = {
            'fecha_inicial': datetime.strptime(values["fecha_inicial"], CALENDAR_FECHA_FMT),
            'fecha_final': datetime.strptime(values["fecha_final"], CALENDAR_FECHA_FMT),
            'rfc_receptor': sat_service.signer.rfc if tipo_recuperar == TipoRecuperar.Recibidas else None,
            'rfc_emisor': sat_service.signer.rfc if tipo_recuperar == TipoRecuperar.Emitidas else None,
            'tipo_solicitud': values["tipo_solicitud"],
        }

        response = sat_service.recover_comprobante_request(
            **args
        )

        self.local_db.solicitud_merge(response["IdSolicitud"], rfc=rfc, request=args, response=response)

    def recupera_comprobantes(self, sat_service, response):
        if response["EstadoSolicitud"] == EstadoSolicitud.TERMINADA:
            for id_paquete in response['IdsPaquetes']:
                r, paquete = sat_service.recover_comprobante_download(
                    id_paquete=id_paquete
                )
                print(f"paquete: {id_paquete}")
                print_yaml(r)
                if paquete:
                    data = base64.b64decode(paquete)
                    with io.BytesIO(data) as b:
                        self.unzip_cfdi(b)

    def unzip_cfdi(self, file):
        title = 'Descomprimiendo'
        with ZipFile(file, "r") as zf:
            try:
                for fileinfo in self.progress_iterate(title, zf.infolist()):
                    data = zf.read(fileinfo)
                    match os.path.splitext(fileinfo.filename)[1]:
                        case ".xml":
                            self._all_invoices = None
                            MyCFDI.move_to_folder(data, pdf_data=None)
                        case ".pdf":
                            pass
                        case ".txt":
                            cfdi_metadata_reader = csv.reader(
                                (c.decode('utf-8') for c in data.splitlines() if c),
                                delimiter='~',
                                quotechar='|'
                            )
                            header = next(cfdi_metadata_reader)
                            for row in cfdi_metadata_reader:
                                row = dict(zip(header, row))
                                print_yaml(row)
                                self.local_db.status_merge(
                                    uuid=row['Uuid'],
                                    estatus=row['Estatus'],
                                    fecha_cancelacion=row['FechaCancelacion']
                                )
            finally:
                self.progress_cancel(title)

    def _read(self, timeout=0):
        event, values = self.window.read(timeout=timeout)
        if event in ("Exit", sg.WIN_CLOSED):
            return False
        return True

    def progress_iterate(self, title, items, fn=None, skip_first=False, delay=0):
        ln = len(items)
        try:
            for i, item in enumerate(items):
                if skip_first and i == 0:
                    pass
                else:
                    if not sg.one_line_progress_meter(
                            '',
                            i,
                            ln,
                            title,
                            fn(item) if callable(fn) else "",
                            key=title,
                            keep_on_top=True,
                            no_titlebar=True,
                            grab_anywhere=True,
                    ):
                        return

                if not self._read(timeout=delay * i):
                    return

                yield item
        finally:
            self.progress_cancel(title)

    @staticmethod
    def progress_cancel(title):
        sg.one_line_progress_meter_cancel(
            key=title
        )

    def action_button(self, action_name, action_items, action_text):
        try:
            match action_name:
                case 'solicitudes':
                    for solicitud in self.progress_iterate(action_text, action_items):
                        rfc = solicitud["rfc"]
                        sat_service = SAT(signer=self.emisores[rfc]['fiel'])

                        id_solicitud = solicitud["response"]["IdSolicitud"]
                        response = sat_service.recover_comprobante_status(
                            id_solicitud=id_solicitud
                        )
                        print_yaml(response)
                        self.local_db.solicitud_merge(id_solicitud, rfc, response=response)
                        self.recupera_comprobantes(sat_service, response)

                case 'facturas' | 'pago':
                    for invoice in self.progress_iterate(action_text, action_items):
                        if not self.generate_invoice(invoice=invoice):
                            break

                case 'correos':
                    with self.email_manager.sender as s:
                        for receptor, facturas, facturas_facturas_pendientes_meses_anteriores in self.progress_iterate(action_text, action_items):
                            def attachments():
                                for ni in facturas:
                                    yield ni.filename + ".xml"
                                    yield ni.filename + ".pdf"

                            s.send_email(
                                subject=f"Comprobantes Fiscales {receptor['RazonSocial']} - {receptor['Rfc']}",
                                to_addrs=receptor["Email"],
                                html=facturacion_environment.get_template('mail_facturas_template.html').render(
                                    facturas=facturas,
                                    facturas_pendientes_meses_anteriores=facturas_facturas_pendientes_meses_anteriores,
                                    receptor=receptor,
                                ),
                                file_attachments=attachments()
                            )
                            for r in facturas:
                                self.local_db.notified_set(r.uuid, True)

                case 'ajustes' | 'depositos':
                    with self.email_manager.sender as s:
                        grouped_action_items = []
                        for _, g_data in itertools.groupby(
                                sorted(
                                    action_items,
                                    key=lambda r: r["receptor"]["Rfc"]
                                ),
                                lambda r: r["receptor"]["Rfc"]
                        ):
                            grouped_action_items.append(list(g_data))

                        for g_data in self.progress_iterate(action_text, grouped_action_items):
                            file_names = []
                            for data in g_data:
                                if file := data['create_fn']():
                                    file_names.append(file)

                            if file_names:
                                receptor = g_data[0]['receptor']
                                if action_name == 'ajustes':
                                    subject = f"Ajuste Renta {receptor['RazonSocial']} - {receptor['Rfc']}"
                                elif action_name == 'depositos':
                                    subject = f"Depósito Renta {receptor['RazonSocial']} - {receptor['Rfc']}"
                                else:
                                    raise NotImplementedError()

                                s.send_email(
                                    subject=subject,
                                    to_addrs=receptor["Email"],
                                    html=facturacion_environment.get_template(f'mail_{action_name}_template.html').render(data),
                                    file_attachments=file_names
                                )

                case 'clientes':
                    for client in self.progress_iterate(
                            action_text, action_items, lambda x: f"Validando: {x['Rfc']}"
                    ):
                        if errors := validar_client(client):
                            self.show_console()
                            for e in errors:
                                print(e)

                case _:
                    raise ValueError(f"Invalid action: {action_name}")
        finally:
            self.progress_cancel(action_text)

    def set_selected_satcfdis_recibidas(self, cfdis: list):
        i = cfdis[0] if len(cfdis) == 1 else None

        if i:
            estatus = EstadoComprobante(i.estatus)
            self.window["status_sat_recibidas"].update(
                estatus.name.center(10),
                disabled=False,
                button_color="red4" if estatus != EstadoComprobante.VIGENTE else "dark green",
            )
        else:
            self.window["status_sat_recibidas"].update(
                "".ljust(10), disabled=True, button_color=sg.theme_background_color()
            )

    def set_selected_satcfdis(self, cfdis: list):
        i = cfdis[0] if len(cfdis) == 1 else None

        if i:
            estatus = EstadoComprobante(i.estatus)
            self.window["status_sat"].update(
                estatus.name.center(10),
                disabled=False,
                button_color="red4" if estatus != EstadoComprobante.VIGENTE else "dark green",
            )
        else:
            self.window["status_sat"].update(
                "".ljust(10), disabled=True, button_color=sg.theme_background_color()
            )

        # Email
        is_active = \
            bool(i) \
            and i["Emisor"]["Rfc"] in self.emisores \
            and i.estatus == EstadoComprobante.VIGENTE
        if is_active:
            self.window["email_notificada"].update(
                "Enviada".center(10) if self.local_db.notified(i) else "Por Enviar",
                disabled=False,
                button_color="dark green" if self.local_db.notified(i) else "red4",
            )
        else:
            self.window["email_notificada"].update(
                "".ljust(10), disabled=True, button_color=sg.theme_background_color()
            )

        # Pendiente de Pago
        is_pendientable = \
            is_active \
            and i["TipoDeComprobante"] == TipoDeComprobante.INGRESO \
            and (i["MetodoPago"] == MetodoPago.PAGO_EN_UNA_SOLA_EXHIBICION or i.saldo_pendiente) \
            and i["Total"]
        if is_pendientable:
            self.window["pendiente_pago"].update(
                (("Pagada" if i["MetodoPago"] == MetodoPago.PAGO_EN_UNA_SOLA_EXHIBICION else "Ignorada")
                 if self.local_db.liquidated(i) else "Por Pagar").center(10),
                disabled=False,
                button_color=("dark green" if i["MetodoPago"] == MetodoPago.PAGO_EN_UNA_SOLA_EXHIBICION else "yellow4")
                if self.local_db.liquidated(i) else "red4",
            )
        else:
            is_ppd_pagada = is_active \
                            and i["TipoDeComprobante"] == TipoDeComprobante.INGRESO \
                            and i["MetodoPago"] == MetodoPago.PAGO_EN_PARCIALIDADES_O_DIFERIDO \
                            and i.saldo_pendiente == 0
            if is_ppd_pagada:
                self.window["pendiente_pago"].update(
                    "Pagada".center(10),
                    disabled=True,
                    button_color="dark green",
                )
            else:
                self.window["pendiente_pago"].update(
                    "".ljust(10),
                    disabled=True,
                    button_color=sg.theme_background_color()
                )

        # PPD
        is_ppd_active = \
            is_active \
            and i["TipoDeComprobante"] == TipoDeComprobante.INGRESO \
            and i["MetodoPago"] == MetodoPago.PAGO_EN_PARCIALIDADES_O_DIFERIDO \
            and i.saldo_pendiente > 0

        self.window["ppd_action_items"].update(visible=is_ppd_active)
        self.window["importe_pago"].update(i.saldo_pendiente if is_ppd_active else '')
        if is_ppd_active:
            self.action_button_manager.set_items("pago", [i])
        else:
            self.action_button_manager.clear()

    def header(self, name, select_console=True):
        if select_console:
            self.show_console()
        self.console.update(header_line(name))
        self._read()

    def show_console(self):
        self.window['errores_tab'].select()

    def download_invoice(self, uuid: UUID):
        res = self.pac_service.recover(uuid, accept=Accept.XML_PDF)
        self._all_invoices = None
        return MyCFDI.move_to_folder(res.xml, pdf_data=res.pdf)

    def facturas_search(self):
        search_text = self.window["emitidas_search"].get()
        search_text = search_text.strip()

        if len(search_text) < 3:
            raise ValueError("Búsqueda debe de tener al menos 3 caracteres")

        def fact_iter():
            if search_text == SearchOptions.PorPagar:
                for i in self.get_all_invoices().values():
                    if i["Emisor"]["Rfc"] in self.emisores \
                            and self.local_db.liquidated_state(i) == StatusState.PENDING:
                        yield i
            elif search_text == SearchOptions.PorEnviar:
                for i in self.get_all_invoices().values():
                    if i["Emisor"]["Rfc"] in self.emisores \
                            and not self.local_db.notified(i) \
                            and i.estatus == EstadoComprobante.VIGENTE:
                        yield i
            elif date_search_text := to_date_period(search_text):
                for i in self.get_all_invoices().values():
                    if i["Emisor"]["Rfc"] in self.emisores \
                            and i["Fecha"] == date_search_text:
                        yield i
            elif uuid_search_text := to_uuid(search_text):
                if uuid_search_text not in self.get_all_invoices():
                    try:
                        self.download_invoice(uuid_search_text)
                    except ResponseError as e:
                        if e.response.status_code == 404:
                            self.error_message(f"Factura no encontrada en el PAC")
                        else:
                            raise e
                if c := self.get_all_invoices().get(uuid_search_text):
                    yield c
            else:
                up_search_text = search_text.upper()
                for i in self.get_all_invoices().values():
                    if i["Emisor"]["Rfc"] in self.emisores \
                            and (
                            i.name == up_search_text
                            or i["Receptor"]["Rfc"] == up_search_text
                            or up_search_text in i["Receptor"].get("Nombre", "")
                    ):
                        yield i

        self.window['emitidas_table'].update(
            values=sorted(fact_iter(), key=lambda x: x["Fecha"], reverse=False),
        )

    def facturas_search_recibidas(self):
        search_text = self.window["recibidas_search"].get()
        search_text = search_text.strip()

        if len(search_text) < 3:
            raise ValueError("Búsqueda debe de tener al menos 3 caracteres")

        def fact_iter():
            if search_text == SearchOptions.PorPagar:
                for i in self.get_all_invoices().values():
                    if i["Receptor"]["Rfc"] in self.emisores \
                            and self.local_db.liquidated_state(i) == StatusState.PENDING:
                        yield i
            elif search_text == SearchOptions.PorEnviar:
                for i in self.get_all_invoices().values():
                    if i["Receptor"]["Rfc"] in self.emisores \
                            and not self.local_db.notified(i) \
                            and i.estatus == EstadoComprobante.VIGENTE:
                        yield i
            elif date_search_text := to_date_period(search_text):
                for i in self.get_all_invoices().values():
                    if i["Receptor"]["Rfc"] in self.emisores \
                            and i["Fecha"] == date_search_text:
                        yield i
            elif uuid_search_text := to_uuid(search_text):
                if uuid_search_text not in self.get_all_invoices():
                    try:
                        self.download_invoice(uuid_search_text)
                    except ResponseError as e:
                        if e.response.status_code == 404:
                            self.error_message(f"Factura no encontrada en el PAC")
                        else:
                            raise e
                if c := self.get_all_invoices().get(uuid_search_text):
                    yield c
            else:
                up_search_text = search_text.upper()
                for i in self.get_all_invoices().values():
                    if i["Receptor"]["Rfc"] in self.emisores \
                            and (
                            i.name == up_search_text
                            or i["Emisor"]["Rfc"] == up_search_text
                            or up_search_text in i["Emisor"].get("Nombre", "")
                    ):
                        yield i

        self.window['recibidas_table'].update(
            values=sorted(fact_iter(), key=lambda x: x["Fecha"], reverse=False),
        )

    def crear_pago(self, values, facturas_pagar):
        clients = ClientsManager()
        fecha_pago = parse_fecha_pago(values["fecha_pago"])

        max_dias = 30
        dif_dias = (datetime.now() - fecha_pago).days
        if dif_dias > max_dias:
            res = sg.popup(
                f"Estas seguro que quieres hacer Complemento de Pago con fecha de pago de hace '{dif_dias}' dias?",
                title="Confirmar",
                button_type=POPUP_BUTTONS_OK_CANCEL,
                location=center_location(self.window),
                keep_on_top=True,
            )
            if res != "OK":
                return []

        importe_pago = parse_importe_pago(values["importe_pago"])
        self.window["importe_pago"].update(importe_pago)

        # noinspection PyUnresolvedReferences
        cfdi = pago_factura(
            receptor_cif=clients[facturas_pagar[0]["Receptor"]["Rfc"]],
            factura_pagar=facturas_pagar[0],
            fecha_pago=fecha_pago,
            forma_pago=values["forma_pago"],
            importe_pago=importe_pago,
            serie_pago=self.local_db.serie_pago(),
        )
        return [cfdi]

    def error_message(self, ex):
        sg.Popup(
            ex,
            no_titlebar=True,
            background_color="red4",
            location=center_location(self.window),
            button_type=POPUP_BUTTONS_NO_BUTTONS,
            auto_close=True,
            non_blocking=True,
        )

    def done_message(self, ex):
        sg.Popup(
            ex,
            no_titlebar=True,
            background_color="green4",
            location=center_location(self.window),
            button_type=POPUP_BUTTONS_NO_BUTTONS,
            auto_close=True,
            non_blocking=True,
            grab_anywhere=True,
        )

    def nuevas_facturas(self, values, force=False):
        facturas_table = self.window['facturas_table']
        has_value = bool(facturas_table.metadata)
        facturas_table.update(values=[])
        self.window['preparar_facturas_text'].update("")

        if has_value or force:
            dp = to_date_period(values["facturas_periodo"])
            if dp is None or dp.month is None:
                raise ValueError("Periodo no válido")

            self.window['preparar_facturas_text'].update(f"{period_desc(dp)}")

            cfdis = generate_ingresos(
                clients=ClientsManager(),
                facturas=FacturasManager(dp)["Facturas"],
                dp=dp
            )
            facturas_table.update(
                values=cfdis,
            )

    def nuevos_ajustes(self, values, force=False):
        ajustes_table = self.window['ajustes_table']
        has_value = bool(ajustes_table.metadata)
        ajustes_table.update(values=[])
        self.window['preparar_ajustes_text'].update("")

        if has_value or force:
            dp_effective = to_date_period(values["ajustes_periodo"])
            if dp_effective is None or dp_effective.month is None:
                raise ValueError("Periodo no válido")

            # dp_effective = add_month(dp, 1)
            self.window['preparar_ajustes_text'].update(f"Ajustes Efectivos Al: {period_desc(dp_effective)}")

            ajustes = generate_ajustes(
                clients=ClientsManager(),
                facturas=FacturasManager(None)["Facturas"],
                dp_effective=dp_effective,
            )
            ajustes_table.update(
                values=ajustes,
            )

    def nuevos_depositos(self, values, force=False):
        depositos_table = self.window['depositos_table']
        has_value = bool(depositos_table.metadata)
        depositos_table.update(values=[])

        if has_value or force:
            depositos = generar_depositos(
                clients=ClientsManager(),
                facturas=FacturasManager(None)["Facturas"],
            )
            depositos_table.update(
                values=depositos,
            )

    def main_tab_group(self, values):
        self.action_button_manager.clear()

        match values['main_tab_group']:
            case 'facturas_tab':
                self.nuevas_facturas(values)

            case 'clientes_tab':
                self.window['clientes_table'].update(
                    values=list(ClientsManager().values()),
                )

            case 'emitidas_tab':
                self.facturas_search()

            case 'recibidas_tab':
                self.facturas_search_recibidas()

            case 'correos_tab':
                now = date.today()
                dp_now = DatePeriod(now.year, now.month)
                clients = ClientsManager()

                def correos():
                    for receptor_rfc, notify_invoices in itertools.groupby(
                            sorted(
                                (i for i in self.get_all_invoices().values()
                                 if i["Emisor"]["Rfc"] in self.emisores
                                    and i.estatus == EstadoComprobante.VIGENTE
                                    and not self.local_db.notified(i)
                                 ),
                                key=lambda r: r["Receptor"]["Rfc"]
                            ),
                            lambda r: r["Receptor"]["Rfc"]
                    ):
                        notify_invoices = list(notify_invoices)

                        def fac_pen_iter():
                            for i in self.get_all_invoices().values():
                                if i["Emisor"]["Rfc"] in self.emisores \
                                        and self.local_db.liquidated_state(i) == StatusState.PENDING \
                                        and i["Fecha"] < dp_now \
                                        and i["Receptor"]["Rfc"] == receptor_rfc \
                                        and i not in notify_invoices:
                                    yield i

                        fac_pen = sorted(
                            fac_pen_iter(),
                            key=lambda r: r["Fecha"]
                        )
                        yield clients[receptor_rfc], notify_invoices, fac_pen

                self.window['correos_table'].update(
                    values=list(correos()),
                )

            case 'ajustes_tab':
                self.nuevos_ajustes(values)

            case 'depositos_tab':
                self.nuevos_depositos(values)

            case 'solicitudes_tab':
                solitudes = self.local_db.get_solicitudes()
                self.window['solicitudes_table'].update(
                    values=list(solitudes.values()),
                )

            case 'contabilidad_tab':
                pass

    def main_loop(self):
        _, values = self.window.read(timeout=0)
        event = "main_tab_group"

        while True:
            if event in ("Exit", sg.WIN_CLOSED):
                return

            self.action(event, values)
            event, values = self.window.read()

    def action(self, event, values):
        try:
            match event:
                case '_focus_in':
                    if not self.has_focus:
                        self.has_focus = True
                        if values["main_tab_group"] in ("clientes_tab", "facturas_tab", "ajustes_tab", "correos_tab"):
                            self.main_tab_group(values)

                case '_focus_out':
                    try:
                        self.has_focus = bool(self.window.TKroot.focus_get())
                    except KeyError:
                        self.has_focus = True

                case "folio_enter":
                    self.set_folio(to_int(values["folio"]))

                case "serie_enter":
                    self.set_serie(values["serie"])

                case "serie_pago_enter":
                    self.set_serie_pago(values["serie_pago"])

                case "about":
                    self.initial_screen()

                case "nueva_solicitud":
                    self.nueva_solicitud(values)
                    self.main_tab_group(values)

                case "buscar_facturas":
                    self.window["emitidas_search"].update(values["buscar_facturas"])
                    self.facturas_search()

                case "buscar_facturas_recibidas":
                    self.window["recibidas_search"].update(values["buscar_facturas_recibidas"])
                    self.facturas_search_recibidas()

                case "emitidas_search_enter":
                    self.facturas_search()

                case "recibidas_search_enter":
                    self.facturas_search_recibidas()

                case "facturas_periodo_enter":
                    self.nuevas_facturas(values, force=True)

                case "ajustes_periodo_enter":
                    self.nuevos_ajustes(values, force=True)

                case 'main_tab_group':
                    self.main_tab_group(values)

                case 'facturas_table+enter':
                    # noinspection PyUnresolvedReferences
                    if s_items := self.window["facturas_table"].selected_items():
                        preview_cfdis(s_items)

                case 'clientes_table+enter':
                    # noinspection PyUnresolvedReferences
                    for client in self.window["clientes_table"].selected_items():
                        url = csf.url(rfc=client["Rfc"], id_cif=client["IdCIF"])
                        open_file(url)

                case 'emitidas_table+enter' | 'recibidas_table+enter':
                    # noinspection PyUnresolvedReferences
                    table = event.split("+")[0]
                    if s_items := self.window[table].selected_items():
                        preview_cfdis(s_items)

                case 'ajustes_table+enter' | 'depositos_table+enter':
                    table = event.split("+")[0]
                    # noinspection PyUnresolvedReferences
                    for items in self.window[table].selected_items():
                        if file_name := items['create_fn']():
                            open_file(
                                os.path.abspath(file_name)
                            )
                        else:
                            self.error_message("No se pudo crear el archivo")

                case 'correos_table+enter' | 'solicitudes_table+enter':
                    pass

                case "facturas_table" | "clientes_table" | "correos_table" | "ajustes_table" | "depositos_table" | "emitidas_table" | "recibidas_table" | "solicitudes_table" | "depositos_table":
                    # noinspection PyUnresolvedReferences
                    s_items = self.window[event].selected_items()
                    if event == "emitidas_table":
                        self.set_selected_satcfdis(s_items)
                    elif event == "recibidas_table":
                        self.set_selected_satcfdis_recibidas(s_items)
                    else:
                        self.action_button_manager.set_items(event.split("_")[0], s_items)

                case "facturas_table+select_all" | "clientes_table+select_all" | "correos_table+select_all" | \
                     "ajustes_table+select_all" | "emitidas_table+select_all" | "recibidas_table" | 'solicitudes_table+select_all' | \
                     'depositos_table+select_all':
                    # noinspection PyUnresolvedReferences
                    self.window[event.split("+")[0]].select_all()

                case "facturas_table+delete" | "clientes_table+delete" | "correos_table+delete" | \
                     "ajustes_table+delete" | "emitidas_table+delete" | "recibidas_table+delete":
                    # noinspection PyUnresolvedReferences
                    # self.window[event.split("+")[0]].delete_selected()
                    pass

                case "solicitudes_table+delete":
                    solitudes = self.local_db.get_solicitudes()
                    # noinspection PyUnresolvedReferences
                    sel = self.window["solicitudes_table"].selected_items()
                    for s in sel:
                        if s["response"].get("EstadoSolicitud").code < EstadoSolicitud.TERMINADA:
                            self.error_message("No se puede eliminar una solicitud en proceso")
                            continue
                        del solitudes[s["response"]["IdSolicitud"]]
                    self.local_db.set_solicitudes(solitudes)
                    self.main_tab_group(values)

                case "status_sat_recibidas":
                    # noinspection PyUnresolvedReferences
                    if i := self.window["recibidas_table"].selected_items()[0]:
                        res = self.local_db.status_sat(i, update=True)
                        self.done_message(f"Estado: {res['Estado']}")
                        self.set_selected_satcfdis_recibidas([i])
                        # noinspection PyUnresolvedReferences
                        self.window['recibidas_table'].refresh()

                case "status_sat":
                    # noinspection PyUnresolvedReferences
                    if i := self.window["emitidas_table"].selected_items()[0]:
                        res = self.local_db.status_sat(i, update=True)
                        # self.done_message(f"Estado: {res['Estado']}")
                        self.done_message(to_yaml(res))
                        self.set_selected_satcfdis([i])
                        # noinspection PyUnresolvedReferences
                        self.window['emitidas_table'].refresh()

                case "pendiente_pago":
                    # noinspection PyUnresolvedReferences
                    if i := self.window["emitidas_table"].selected_items()[0]:
                        self.local_db.liquidated_flip(i)
                        self.set_selected_satcfdis([i])
                        # noinspection PyUnresolvedReferences
                        self.window['emitidas_table'].refresh()

                case "email_notificada":
                    # noinspection PyUnresolvedReferences
                    if i := self.window["emitidas_table"].selected_items()[0]:
                        self.local_db.notified_flip(i)
                        self.set_selected_satcfdis([i])
                        # noinspection PyUnresolvedReferences
                        self.window['emitidas_table'].refresh()

                case "crear_facturas" | "ver_preview":
                    action_text = self.action_button_manager.text()
                    action_name = self.action_button_manager.name
                    action_items = self.action_button_manager.items

                    if action_name == "pago":
                        action_items = self.crear_pago(values, action_items)
                        if not action_items:
                            return

                    if event == "ver_preview":
                        if action_name in ("facturas", "pago"):
                            preview_cfdis(action_items)

                    elif event == "crear_facturas":
                        res = sg.popup(
                            f"Estas seguro que quieres '{action_text}'?",
                            title="Confirmar",
                            button_type=POPUP_BUTTONS_OK_CANCEL,
                            location=center_location(self.window),
                            keep_on_top=True,
                        )
                        if res == "OK":
                            self.header(action_text.upper(), select_console=False)
                            self.action_button_manager.clear()
                            self.action_button(
                                action_name=action_name,
                                action_items=action_items,
                                action_text=action_text
                            )
                            self.main_tab_group(values)

                case "editar_clientes":
                    open_file(
                        os.path.abspath("clientes.yaml")
                    )

                case "exportar_clientes":
                    filename = 'clientes.txt'
                    clients = ClientsManager()
                    clientes_generar_txt(filename, clients)
                    open_file(
                        os.path.dirname(
                            os.path.abspath(filename)
                        )
                    )

                case "editar_facturas" | "editar_ajustes":
                    open_file(
                        os.path.abspath("facturas.yaml")
                    )

                case "editar_configurar":
                    open_file(
                        os.path.abspath("config.yaml")
                    )

                case "ver_config":
                    open_file(
                        os.path.abspath(".")
                    )

                case "ver_excel":
                    rfc = values["contabilidad_rfc"]
                    dp = to_date_period(values["periodo"])

                    archivo_excel = exportar_facturas(
                        self.get_all_invoices(),
                        dp,
                        ClientsManager()[rfc],
                        self.rfc_prediales
                    )
                    open_file(
                        os.path.abspath(archivo_excel)
                    )

                case "ver_diot":
                    rfc = values["contabilidad_rfc"]
                    dp = to_date_period(values["periodo"])

                    archivo_diot = calculate_diot(
                        self.get_all_invoices(),
                        dp,
                        ClientsManager()[rfc],
                    )
                    open_file(
                        os.path.abspath(archivo_diot)
                    )

                case "periodo_enter":
                    rfc = values["contabilidad_rfc"]
                    dp = to_date_period(values["periodo"])

                    declaracion_provisional = calculate_declaracion_provisional(
                        self.get_all_invoices(),
                        dp,
                        ClientsManager()[rfc],
                        self.rfc_prediales
                    )
                    self.window['declaracion'].update(
                        append=False,
                        value=declaracion_provisional
                    )

                case "ver_carpeta":
                    dp = to_date_period(values["periodo"])
                    directory = archivos_folder(dp)
                    open_file(
                        os.path.abspath(directory)
                    )

                case "organizar_facturas":
                    MyCFDI.rename_invoices(search_path="**/*.xml")

                case "cargar_zip":
                    zip_file = sg.popup_get_file('', multiple_files=False, no_window=True, file_types=(("ZIP Files", "*.zip"),))
                    if zip_file:
                        self.unzip_cfdi(zip_file)

                case 'importar_emitidas':
                    csv_file = sg.popup_get_file('', multiple_files=False, no_window=True, file_types=(("CSV Files", "*.csv"),))
                    if csv_file:
                        all_invoices = self.get_all_invoices()
                        with open(csv_file, newline='', encoding='utf-8') as f:
                            reader = csv.reader(f)
                            header = next(reader)
                            for row in reader:
                                row = dict(zip(header, row))
                                uuid = UUID(row["Folio Fiscal (UUID)"])
                                if uuid not in all_invoices:
                                    cfdi = self.download_invoice(uuid)

                                    if row.get("Estatus", "Entregado SAT") != "Entregado SAT":
                                        self.local_db.status_sat(cfdi, update=True)
                        self.done_message("FIN")

                case "exportar_metadata":
                    with open(METADATA_FILE, 'w', newline='', encoding='utf-8') as f:
                        writer = csv.writer(f)
                        for i in self.get_all_invoices():
                            if status := self.local_db.status_export(i):
                                writer.writerow(status)
                    self.done_message("FIN")

                case "importar_metadata":
                    with open(METADATA_FILE, newline='', encoding='utf-8') as f:
                        reader = csv.reader(f)
                        for row in reader:
                            self.local_db.status_merge(*row)
                    self.done_message("FIN")

                case _:
                    logger.error(f"Unknown event '{event}'")

        except (ValueError, XlsxFileError) as ex:
            self.error_message(ex)
        except ConsoleErrors as ex:
            self.header(str(ex))
            for error in ex.errors:
                self.console.update(f"{error}\n", append=True)
        except Exception as ex:
            self.header("Exception")
            self.console.update(append=True, value=str(ex))
            logger.exception("Main Loop Exception")



import logging
import os
from zipfile import ZipFile

import PySimpleGUI as sg

SOURCE_DIRECTORY = os.path.dirname(__file__)

DATA_DIRECTORY = ".data"
ARCHIVOS_DIRECTORY = "archivos"
METADATA_FILE = os.path.join(ARCHIVOS_DIRECTORY, "metadata.csv")
TEMPLATES_DIRECTORY = "templates"
TEMP_DIRECTORY = ".data/temp"


def add_file_handler():
    os.makedirs(DATA_DIRECTORY, exist_ok=True)
    fh = logging.FileHandler(
        os.path.join(DATA_DIRECTORY, 'errors.log'),
        mode='a',
        encoding='utf-8',
    )
    fh.setLevel(logging.ERROR)
    formatter = logging.Formatter('%(asctime)s - %(message)s')
    fh.setFormatter(formatter)
    logging.root.addHandler(fh)


class FacturacionLauncher:
    def __init__(self, cwd=None):
        # sg.theme('Reddit')
        self.app_dir = os.getcwd()
        if cwd:
            os.chdir(cwd)
        add_file_handler()

        # layout
        layout = [
            [sg.Column([[
                sg.Image(source=os.path.join(SOURCE_DIRECTORY, "images", "logo.png"), pad=2),
            ]], justification='center', pad=0, background_color='black')],
            [
                sg.Multiline(
                    "Cargando Aplicación...",
                    expand_x=True,
                    expand_y=True,
                    key="console",
                    background_color='#a0dbd9',
                    no_scrollbar=True,
                    border_width=0,
                    write_only=True,
                    disabled=True,
                )
            ]
        ]

        self.window = sg.Window(
            f"Facturación Masiva CFDI 4.0",
            layout,
            size=(640, 480),
            resizable=True,
            font=("Courier New", 11, "bold"),
            no_titlebar=True,
            modal=True,
            background_color='#a0dbd9',  # sg.theme_background_color(),
            auto_close=True,
            auto_close_duration=10,  # seconds
        )

    @staticmethod
    def read_config():
        from satdigitalinvoice.file_data_managers import ConfigManager
        return ConfigManager()

    def add_sample_files(self):
        # loading the sample.zip
        try:
            with ZipFile(os.path.join(self.app_dir, 'sample.zip'), 'r') as zf:
                for member in zf.infolist():
                    if not os.path.exists(member.filename):
                        zf.extract(member)
        except FileNotFoundError:
            pass

    def run(self):
        self.window.finalize()
        self.window.read(timeout=0)

        try:
            # check if another directory is configured
            from satdigitalinvoice.file_data_managers import InitManager
            if cwd := InitManager().get('cwd'):
                os.chdir(cwd)
            self.add_sample_files()

            from satdigitalinvoice.facturacion import FacturacionGUI
            app = FacturacionGUI(
                config=self.read_config()
            )
        except Exception as ex:
            logging.exception(ex)
            self.window['console'].update(
                f"Error al cargar la aplicación. {ex.__class__.__name__}: {ex}",
                text_color='red4',
            )
            self.window.read(close=True)
            return

        self.window.close()
        app.run()



from satdigitalinvoice.formatting_functions.common import pesos
from decimal import Decimal


def test_pesos():
    assert pesos(1000.0) == '$1,000.00 (SON: MIL PESOS 00/100M.N.)'
    assert pesos(Decimal("123.123")) == '$123.12 (SON: CIENTO VEINTITRÉS PESOS 12/100M.N.)'



def get_invoices_data(all_invoices, dp, rfc):
    for c in all_invoices.values():
        if c["Fecha"] == dp and c["Emisor"]["Rfc"] == rfc and c["TipoDeComprobante"] == "I":
            # print(c)
            print("- Rfc:", c["Receptor"]["Rfc"])
            print("  UsoCFDI:", c["Receptor"]["UsoCFDI"].code)
            print("  MetodoPago:", c["MetodoPago"].code)
            print("  Conceptos:")
            print("  - CuentaPredial:", repr(c["Conceptos"][0]["CuentaPredial"]))
            print("    ClaveProdServ:", repr(c["Conceptos"][0]["ClaveProdServ"].code))
            print("    Cantidad:", "1.0")
            print("    ClaveUnidad:", c["Conceptos"][0]["ClaveUnidad"].code)
            print("    Descripcion:", repr(c["Conceptos"][0]["Descripcion"].replace("MES DE SEPTIEMBRE DEL 2022", "{periodo}")))
            print("    ValorUnitario:", c["Conceptos"][0]["ValorUnitario"])
            print("  Total:", c["Total"])
            print()



from decimal import Decimal

from satcfdi.create.cfd.catalogos import Impuesto, TipoFactor, RegimenFiscal
from satcfdi.models import DatePeriod, RFC, RFCType
from satcfdi.utils import iterate

ISR_MENSUAL_2023 = [  # limite, cuota_fija, porcentaje
    (Decimal('375975.62'), Decimal("117912.32"), Decimal("0.3500")),
    (Decimal('125325.21'), Decimal("32691.18"), Decimal("0.3400")),
    (Decimal('93993.91'), Decimal("22665.17"), Decimal("0.3200")),
    (Decimal('49233.01'), Decimal("9236.89"), Decimal("0.3000")),
    (Decimal('31236.50'), Decimal("5004.12"), Decimal("0.2352")),
    (Decimal('15487.72'), Decimal("1640.18"), Decimal("0.2136")),
    (Decimal('12935.83'), Decimal("1182.88"), Decimal("0.1792")),
    (Decimal('11128.02'), Decimal("893.63"), Decimal("0.1600")),
    (Decimal('6332.06'), Decimal("371.83"), Decimal("0.1088")),
    (Decimal('746.05'), Decimal("14.32"), Decimal("0.0640")),
    (Decimal("0.00"), Decimal("0.00"), Decimal("0.0192")),
]

ISR_ANUAL_2023 = [  # limite, cuota_fija, porcentaje
    (Decimal('4511707.38'), Decimal("1414947.85"), Decimal("0.3500")),
    (Decimal('1503902.47'), Decimal("392294.17"), Decimal("0.3400")),
    (Decimal('1127926.85'), Decimal("271981.99"), Decimal("0.3200")),
    (Decimal('590796.00'), Decimal("110842.74"), Decimal("0.3000")),
    (Decimal('374837.89'), Decimal("60049.40"), Decimal("0.2352")),
    (Decimal('185852.58'), Decimal("19682.13"), Decimal("0.2136")),
    (Decimal('155229.81'), Decimal("14194.54"), Decimal("0.1792")),
    (Decimal('133536.08'), Decimal("10723.55"), Decimal("0.1600")),
    (Decimal('75984.56'), Decimal("4461.94"), Decimal("0.1088")),
    (Decimal('8952.50'), Decimal("171.88"), Decimal("0.0640")),
    (Decimal("0.00"), Decimal("0.00"), Decimal("0.0192")),
]

ISR_MENSUAL_RESICO_2023 = [
    (Decimal("208333.34"), None, Decimal("0.0250")),
    (Decimal("83333.34"), None, Decimal("0.0200")),
    (Decimal("50000.01"), None, Decimal("0.0150")),
    (Decimal("25000.01"), None, Decimal("0.0110")),
    (Decimal("0.00"), None, Decimal("0.0100")),
]


def isr_mensual(dp: DatePeriod, ingreso):
    if dp.month is None:
        isr_table = ISR_ANUAL_2023
    else:
        isr_table = ISR_MENSUAL_2023

    for (limite, cuota_fija, porcentaje) in isr_table:
        if ingreso >= limite:
            return round((ingreso - limite) * porcentaje + cuota_fija)
    return Decimal("0.00")


def sat_retenciones(concepto, emisor, receptor):
    if 'Retenciones' in concepto['Impuestos']:
        return

    emisor_rfc = RFC(emisor['Rfc'])
    receptor_rfc = RFC(receptor['Rfc'])

    if emisor_rfc.type == RFCType.FISICA and receptor_rfc.type == RFCType.MORAL:
        ret_isr = {
            'Impuesto': Impuesto.ISR,
            'TipoFactor': TipoFactor.TASA,
            'TasaOCuota': Decimal('0.100000') if emisor['RegimenFiscal'] != RegimenFiscal.REGIMEN_SIMPLIFICADO_DE_CONFIANZA else Decimal('0.012500')
        }

        if receptor['RegimenFiscal'] == '603':  # Personas morales con fines no lucrativos
            concepto['Impuestos']['Retenciones'] = ret_isr  # solo se retiene ISR
        else:
            traslado = next(i for i in iterate(concepto['Impuestos']['Traslados']) if i['Impuesto'] == Impuesto.IVA)
            assert traslado is not None
            assert traslado['TipoFactor'] == TipoFactor.TASA

            concepto['Impuestos']['Retenciones'] = [
                ret_isr,
                {
                    'Impuesto': Impuesto.IVA,
                    'TipoFactor': TipoFactor.TASA,
                    'TasaOCuota': round(traslado['TasaOCuota'] * 2 / 3, 6)
                }
            ]



import copy
import logging
import os
from datetime import datetime, date
from decimal import Decimal
from decimal import InvalidOperation
from itertools import groupby

import xlsxwriter
from markdown2 import markdown
from satcfdi import render
from satcfdi.accounting import filter_invoices_iter, filter_payments_iter, invoices_export, payments_export
from satcfdi.accounting.process import payments_groupby_receptor, payments_retentions_export
from satcfdi.create.cfd import cfdi40
from satcfdi.create.cfd.catalogos import Impuesto
from satcfdi.create.cfd.cfdi40 import Comprobante, PagoComprobante
from satcfdi.diot import DIOT, DatosIdentificacion, Periodo, ProveedorTercero, TipoTercero, TipoOperacion
from satcfdi.models import DatePeriod
from satcfdi.pacs import sat

try:
    from weasyprint import HTML, CSS
except:
    pass

from . import SOURCE_DIRECTORY, ARCHIVOS_DIRECTORY, TEMP_DIRECTORY
from .environments import facturacion_environment
from .exceptions import ConsoleErrors
from .formatting_functions.common import fecha_mes, get_month_name
from .log_tools import to_yaml
from .utils import add_month, find_best_match, months_between, open_file
from .sat_functions import isr_mensual, sat_retenciones

logger = logging.getLogger(__name__)
logger.level = logging.INFO

sat_manager = sat.SAT()

PERIODICIDAD = {
    "Mensual": 1,
    "Bimestral": 2,
    "Trimestral": 3,
    "Cuatrimestral": 4,
    "Semestral": 6,
    "Anual": 12
}
CALENDAR_FECHA_FMT = '%Y-%m-%d'


def create_cfdi(emisor_cif, receptor_cif, factura_details):
    invoice = cfdi40.Comprobante(
        emisor=cfdi40.Emisor(
            rfc=emisor_cif['Rfc'],
            nombre=emisor_cif['RazonSocial'],
            regimen_fiscal=factura_details.get("EmisorRegimen") or emisor_cif['RegimenFiscal']
        ),
        lugar_expedicion=emisor_cif['CodigoPostal'],
        receptor=cfdi40.Receptor(
            rfc=receptor_cif['Rfc'],
            nombre=receptor_cif['RazonSocial'],
            uso_cfdi=factura_details['UsoCFDI'],
            domicilio_fiscal_receptor=receptor_cif['CodigoPostal'],
            regimen_fiscal_receptor=receptor_cif['RegimenFiscal']
        ),
        metodo_pago=factura_details['MetodoPago'],
        forma_pago=factura_details['FormaPago'],
        # cfdi_relacionados=[
        #     cfdi40.CfdiRelacionados(
        #         tipo_relacion=TipoRelacion.SUSTITUCION_DE_LOS_CFDI_PREVIOSaaaaaaa,
        #         cfdi_relacionado=''xyc',
        #     )
        # ],
        # serie=serie,
        # folio=folio,
        conceptos=factura_details["Conceptos"]
    )
    invoice = invoice.process()
    return invoice


def parse_periodo_mes_ajuste(periodo_mes_ajuste: str):
    parts = periodo_mes_ajuste.split(".")
    if len(parts) != 2:
        raise ValueError("Periodo Invalido")

    periodo, mes_ajuste = parts
    if not mes_ajuste.isnumeric():
        raise ValueError("Periodo Invalido")

    mes_ajuste = int(mes_ajuste)
    if not (12 >= int(mes_ajuste) >= 1):
        raise ValueError("Periodo Invalido")

    if periodo not in PERIODICIDAD:
        raise ValueError("Periodo Invalido")

    return PERIODICIDAD[periodo], mes_ajuste


def format_concepto_desc(concepto, periodo):
    template = facturacion_environment.from_string(concepto["Descripcion"])
    concepto["Descripcion"] = template.render(
        periodo=periodo
    )
    return concepto


def period_desc(dp: DatePeriod):
    if dp.month:
        return fecha_mes(date(year=dp.year, month=dp.month, day=1))
    return f"AÑO {dp.year}"


def periodicidad_mes_desc(periodo_mes_ajuste):
    _, mes_ajuste = parse_periodo_mes_ajuste(periodo_mes_ajuste)
    return get_month_name(mes_ajuste)


def periodicidad_desc(dp: DatePeriod, periodo_mes_ajuste, offset):
    periodo_meses, mes_ajuste = parse_periodo_mes_ajuste(periodo_mes_ajuste)

    if (dp.month - mes_ajuste) % periodo_meses == 0:
        if offset:
            dp = add_month(dp, offset)

        periodo = period_desc(dp)
        if periodo_meses > 1:
            periodo += " AL " + period_desc(
                dp=add_month(dp, periodo_meses - 1)
            )

        return periodo
    return None


def generate_ingresos(clients, facturas, dp):
    errors = []

    def facturas_iter():
        for i, f in enumerate(facturas, start=1):
            try:
                emisor_cif = clients.get(f['Emisor'])
                if not emisor_cif:
                    raise ValueError("Emisor not found")

                receptor_cif = clients.get(f['Receptor'])
                if not receptor_cif:
                    raise ValueError("Receptor not found")

                def prepare_concepto(c):
                    periodo = periodicidad_desc(
                        dp,
                        c['_periodo_mes_ajuste'],
                        c.get('_desfase_mes')
                    )
                    if periodo and c['ValorUnitario'] is not None:
                        c = copy.deepcopy(c)
                        sat_retenciones(c, emisor_cif, receptor_cif)
                        return format_concepto_desc(c, periodo=periodo)

                if f["MetodoPago"] == "PPD" and f["FormaPago"] != "99":
                    raise ValueError(f"FormaPago '{f['FormaPago']}' is invalid, expected '99' for PPD")

                if conceptos := [x for x in (prepare_concepto(c) for c in f["Conceptos"]) if x]:
                    f["Conceptos"] = conceptos
                    cfdi = create_cfdi(emisor_cif, receptor_cif, f)

                    expected_total = f.get('Total')
                    if expected_total is not None and expected_total != cfdi['Total']:
                        raise ValueError(f"Total '{expected_total}' is invalid, expected '{cfdi['Total']}'")

                    yield cfdi
            except Exception as e:
                errors.append(f"{i} {f['Receptor']}: {str(e)}")

    cfdis = list(facturas_iter())
    if errors:
        raise ConsoleErrors("Generar Facturas Errores", errors=errors)
    return cfdis


def parse_fecha_pago(fecha_pago):
    if not fecha_pago:
        raise ValueError("Fecha de Pago es requerida")

    fecha_pago = datetime.strptime(fecha_pago, CALENDAR_FECHA_FMT)
    if fecha_pago > datetime.now():
        raise ValueError("Fecha de Pago es mayor a la fecha actual")

    if fecha_pago.replace(hour=12) > datetime.now():
        fecha_pago = datetime.now()
    else:
        fecha_pago = fecha_pago.replace(hour=12)

    return fecha_pago


def parse_importe_pago(importe_pago: str):
    try:
        return round(Decimal(importe_pago), 2)
    except InvalidOperation:
        raise ValueError("Importe de Pago es invalido")


def pago_factura(factura_pagar, fecha_pago: datetime, forma_pago: str, importe_pago: Decimal = None, serie_pago=None, receptor_cif=None):
    c = factura_pagar

    receptor = cfdi40.Receptor(
        rfc=receptor_cif['Rfc'],
        nombre=receptor_cif['RazonSocial'],
        domicilio_fiscal_receptor=receptor_cif['CodigoPostal'],
        regimen_fiscal_receptor=receptor_cif['RegimenFiscal'],
        uso_cfdi="CP01"
    )

    invoice = Comprobante.pago_comprobantes(
        receptor=receptor,
        serie=serie_pago or None,
        folio=c.get('Folio') if serie_pago else None,
        comprobantes=[
            PagoComprobante(
                comprobante=c,
                num_parcialidad=c.ultima_num_parcialidad + 1,
                imp_saldo_ant=c.saldo_pendiente,
                imp_pagado=importe_pago
            )
        ],
        fecha_pago=fecha_pago,
        forma_pago=forma_pago,
    )
    return invoice.process()


def iter_conceptos(facturas):
    for f in facturas:
        rfc_emisor = f["Emisor"]
        rfc_receptor = f["Receptor"]
        for c in f["Conceptos"]:
            yield rfc_emisor, rfc_receptor, c


def find_ajustes(facturas, mes_ajuste):
    for rfc_emisor, rfc_receptor, concepto in iter_conceptos(facturas):
        _, mes_aj = parse_periodo_mes_ajuste(concepto['_periodo_mes_ajuste'])
        if mes_aj == mes_ajuste:
            yield rfc_emisor, rfc_receptor, concepto


def generate_ajustes(clients, facturas, dp_effective):
    errors = []
    ajustes_dir = os.path.join(archivos_folder(dp_effective), 'ajustes')
    os.makedirs(ajustes_dir, exist_ok=True)

    def ajustes_iter():
        for i, (emisor_rfc, receptor_rfc, concepto) in enumerate(find_ajustes(facturas, dp_effective.month), start=1):
            # try:
            valor_unitario_raw = concepto["ValorUnitario"]

            if isinstance(valor_unitario_raw, dict):
                vu_eff, vu = find_best_match(valor_unitario_raw, add_month(dp_effective, -1))
                vun_eff, vun = find_best_match(valor_unitario_raw, dp_effective)
                if vu_eff == vun_eff or vun is None or vu is None:
                    vun = None
                    num_meses = None
                else:
                    num_meses = months_between(vun_eff, vu_eff)

                if vun and vu:
                    ajuste_porcentaje = round((vun / vu - 1) * 100, 2)
                else:
                    ajuste_porcentaje = None
            else:
                vu = valor_unitario_raw
                vun = None
                num_meses = None
                ajuste_porcentaje = None

            concepto = format_concepto_desc(concepto, periodo="INMUEBLE")
            file_name = os.path.join(ajustes_dir, f'AjusteRenta_{receptor_rfc}_{i}.pdf')
            data = {
                "receptor": clients[receptor_rfc],
                "emisor": clients[emisor_rfc],
                "concepto": concepto,
                "valor_unitario": vu,
                "valor_unitario_nuevo": vun or '',
                "ajuste_porcentaje": ajuste_porcentaje or "",
                "meses": num_meses or '',
                "efectivo_periodo_desc": periodicidad_desc(dp_effective, concepto['_periodo_mes_ajuste'], concepto.get('_desfase_mes')),
                "periodo": concepto['_periodo_mes_ajuste'].split('.')[0].upper(),
            }
            data['create_fn'] = create_ajuste_fn(ajuste_porcentaje, data, file_name)

            yield data
        # except Exception as e:
        #     errors.append(f"{i} {receptor_rfc}: {str(e)}")

    cfdis = list(ajustes_iter())
    if errors:
        raise ConsoleErrors("Generar Ajustes Errores", errors=errors)
    return cfdis


def create_ajuste_fn(ajuste_porcentaje, data, file_name):
    def fn():
        if ajuste_porcentaje:
            generate_pdf_template(
                template_name='ajuste.md',
                fields=data,
                target=file_name,
            )
            return file_name

    return fn


def find_depositos(facturas):
    for rfc_emisor, rfc_receptor, concepto in iter_conceptos(facturas):
        if "_deposito" in concepto:
            yield rfc_emisor, rfc_receptor, concepto


def generar_depositos(clients, facturas):
    errors = []
    depositos_dir = os.path.join(ARCHIVOS_DIRECTORY, 'depositos')
    os.makedirs(depositos_dir, exist_ok=True)

    def depositos_iter():
        for i, (emisor_rfc, receptor_rfc, concepto) in enumerate(find_depositos(facturas), start=1):
            try:
                dep = concepto["_deposito"]

                concepto = format_concepto_desc(concepto, periodo="INMUEBLE")
                file_name = os.path.join(depositos_dir, f'Deposito_{receptor_rfc}_{i}.pdf')

                data = {
                    "receptor": clients[receptor_rfc],
                    "emisor": clients[emisor_rfc],
                    "concepto": concepto,
                    "deposito": dep,
                    "periodo": concepto['_periodo_mes_ajuste'].split('.')[0].upper(),
                }
                data['create_fn'] = create_deposito_fn(data, file_name)
                yield data
            except Exception as e:
                errors.append(f"{i} {receptor_rfc}: {str(e)}")

    cfdis = list(depositos_iter())
    if errors:
        raise ConsoleErrors("Generar Depositos Errores", errors=errors)
    return cfdis


def create_deposito_fn(data, file_name):
    def fn():
        generate_pdf_template(
            template_name='deposito.md',
            fields=data,
            target=file_name,
        )
        return file_name

    return fn


def archivos_folder(dp: DatePeriod):
    if dp.month:
        return os.path.join(ARCHIVOS_DIRECTORY, str(dp.year), str(dp.year) + "-{:02d}".format(dp.month))
    return os.path.join(ARCHIVOS_DIRECTORY, str(dp.year))


def archivos_filename(dp: DatePeriod, name=".xlsx"):
    return os.path.join(archivos_folder(dp), f"{dp}{name}")


def exportar_facturas(all_invoices, dp: DatePeriod, emisor_cif, rfc_prediales):
    emisor_rfc = emisor_cif['Rfc']

    emitidas = filter_invoices_iter(invoices=all_invoices.values(), fecha=dp, rfc_emisor=emisor_rfc)
    emitidas_pagos = filter_payments_iter(invoices=all_invoices, fecha=dp, rfc_emisor=emisor_rfc)
    emitidas_pagos = list(emitidas_pagos)

    recibidas = filter_invoices_iter(invoices=all_invoices.values(), fecha=dp, rfc_receptor=emisor_rfc)
    recibidas_pagos = filter_payments_iter(invoices=all_invoices, fecha=dp, rfc_receptor=emisor_rfc)

    recibidas_pagos = list(recibidas_pagos)
    pagos_hechos_iva = [
        p
        for p in recibidas_pagos
        if sum(x.get("Importe", 0) for x in p.impuestos.get("Traslados", {}).values()) > 0 and p.comprobante["Receptor"].get("RegimenFiscalReceptor") not in ('616',)
    ]
    prediales = [p for p in recibidas_pagos if p.comprobante["Emisor"]["Rfc"] in rfc_prediales]

    archivo_excel = archivos_filename(dp)
    os.makedirs(os.path.dirname(archivo_excel), exist_ok=True)

    workbook = xlsxwriter.Workbook(archivo_excel)
    # EMITIDAS
    invoices_export(workbook, "EMITIDAS", emitidas)
    payments_export(workbook, "EMITIDAS PAGOS", emitidas_pagos)

    # RECIBIDAS
    invoices_export(workbook, "RECIBIDAS", recibidas)
    payments_export(workbook, "RECIBIDAS PAGOS", recibidas_pagos)

    # SPECIALES
    payments_export(workbook, f"RECIBIDAS PAGOS IVA", pagos_hechos_iva)
    if prediales:
        payments_export(workbook, "PREDIALES", prediales)

    # RETENCIONES
    if dp.month is None:
        archivo_retenciones = archivos_filename(dp, name="retenciones.txt")
        pagos_agrupados = payments_groupby_receptor(emitidas_pagos)
        payments_retentions_export(archivo_retenciones, pagos_agrupados)

    workbook.close()
    return archivo_excel


def calculate_declaracion_provisional(all_invoices, dp: DatePeriod, emisor_cif, rfc_prediales):
    emisor_rfc = emisor_cif['Rfc']

    emitidas_pagos = list(filter_payments_iter(invoices=all_invoices, fecha=dp, rfc_emisor=emisor_rfc))
    recibidas_pagos = list(filter_payments_iter(invoices=all_invoices, fecha=dp, rfc_receptor=emisor_rfc))
    prediales = [p for p in recibidas_pagos if p.comprobante["Emisor"]["Rfc"] in rfc_prediales]

    emitidas_pagos = sum_payments(emitidas_pagos)
    recibidas_pagos = sum_payments(recibidas_pagos)
    prediales = sum_payments(prediales)['Subtotal']

    # ISR
    total_ingresos = emitidas_pagos['Subtotal']
    deduccion_opcional = round(total_ingresos * Decimal("0.35"))
    total_deducciones = deduccion_opcional + prediales
    base_gravable = total_ingresos - deduccion_opcional - prediales
    isr_causado = isr_mensual(dp, base_gravable)
    isr_a_cargo = isr_causado - emitidas_pagos['ISR Ret']

    # IVA
    iva_a_cargo = round(total_ingresos * Decimal("0.16")) - emitidas_pagos['IVA Ret'] - recibidas_pagos['IVA16 Tras']
    res = {
        "ISR PERSONAS FÍSICAS. ARRENDAMIENTO DE INMUEBLES (USO O GOCE)": {
            "¿Tus ingresos fueron obtenidos en copropiedad o sociedad conyugal?": "No",
            "Tipo de deducción": "Deducción opcional",
            "Total de ingresos": total_ingresos,
            "Deducción opcional": deduccion_opcional,
            "Impuesto predial": prediales,
            "Total de deducciones autorizadas": total_deducciones,
            "Base gravable del pago provisional": base_gravable,
            "ISR causado": isr_causado,
            "Estímulos acreditables": 0,
            "Impuesto retenido": emitidas_pagos['ISR Ret'],
            "ISR a cargo": isr_a_cargo,
        },
        "IMPUESTO AL VALOR AGREGADO": {
            "Actividades gravadas a la tasa del 16%": total_ingresos,
            "Actividades gravadas a la tasa del 0%": 0,
            "Actividades exentas": 0,
            "Actividades no objeto del impuesto": 0,
            "IVA cobrado del periodo a la tasa del 16%": round(total_ingresos * Decimal("0.16")),
            "IVA acreditable del periodo": recibidas_pagos['IVA16 Tras'],
            "IVA retenido": emitidas_pagos['IVA Ret'],
            "¿Tienes otras cantidades a cargo?": 'No',
            "Cantidad a cargo": iva_a_cargo,
            "¿Tienes otras cantidades a favor?": 'No',
            "Impuesto a cargo": iva_a_cargo,
        },
        "TOTAL": isr_a_cargo + iva_a_cargo

    }
    p_desc = period_desc(dp)
    return p_desc + "\n" + emisor_rfc + "\n" + to_yaml(res)


def calculate_diot(all_invoices, dp: DatePeriod, emisor_cif):
    emisor_rfc = emisor_cif['Rfc']
    recibidas_pagos = filter_payments_iter(invoices=all_invoices, fecha=dp, rfc_receptor=emisor_rfc)
    recibidas_pagos = list(r for r in recibidas_pagos if r.comprobante["Receptor"].get("RegimenFiscalReceptor") not in ('616',))
    recibidas_pagos.sort(key=lambda x: x.comprobante["Emisor"]["Rfc"])

    provedores = {}
    for rfc, group in groupby(recibidas_pagos, lambda x: x.comprobante["Emisor"]["Rfc"]):
        payments = list(group)
        provedores[rfc] = {
            'Subtotal': sum(i.sub_total for i in payments),
            "002|Tasa|0.160000": sum(i.impuestos.get("Traslados", {}).get("002|Tasa|0.160000", {}).get("Importe", 0) for i in payments),
        }

    diot = DIOT(
        datos_identificacion=DatosIdentificacion(
            rfc=emisor_rfc,
            curp=emisor_cif['CURP'],
            nombre=emisor_cif['Nombre'],
            apellido_paterno=emisor_cif['ApellidoPaterno'],
            apellido_materno=emisor_cif['ApellidoMaterno'],
            ejercicio=2021,
        ),
        periodo=f'{dp.month:02d}',
        proveedores=[
            ProveedorTercero(
                tipo_tercero=TipoTercero.PROVEEDOR_NACIONAL,
                tipo_operacion=TipoOperacion.OTROS,
                rfc=rfc,
                iva16=round(values["002|Tasa|0.160000"] / Decimal("0.16")),
            )
            for rfc, values in provedores.items() if values["002|Tasa|0.160000"]
        ]
    )

    diot.generate_package(
        dirname=archivos_folder(dp)
    )

    diot_pdf = archivos_filename(dp, name="diot.pdf")
    render.pdf_write(diot, diot_pdf)

    diot_export = archivos_filename(dp, name="diot.txt")
    with open(diot_export, "wb") as f:
        diot.export(f)
    return diot_pdf


def sum_payments(payments):
    return {
        'Subtotal': round(sum(i.sub_total for i in payments)),
        'ISR Ret': round(sum(i.impuestos.get("Retenciones", {}).get(Impuesto.ISR, {}).get("Importe", 0) for i in payments)),
        'IVA Ret': round(sum(i.impuestos.get("Retenciones", {}).get(Impuesto.IVA, {}).get("Importe", 0) for i in payments)),

        'IVA16 Tras': round(sum(i.impuestos.get("Traslados", {}).get("002|Tasa|0.160000", {}).get("Importe", 0) for i in payments)),
    }


def generate_pdf_template(template_name, fields, target=None, css_string=None):
    template = facturacion_environment.get_template(template_name)
    md5_document = template.render(
        fields
    )
    html = markdown(md5_document)
    pdf = HTML(string=html).write_pdf(
        target=target,
        stylesheets=[
            os.path.join(SOURCE_DIRECTORY, "markdown_styles", "markdown6.css"),
            CSS(
                string=css_string or '@page { width: Letter; margin: 1.6cm 1.6cm 1.6cm 1.6cm; }'
            )
        ]
    )
    return pdf


def mf_pago_fmt(cfdi):
    i = cfdi
    if i['TipoDeComprobante'] == "I":
        return i['TipoDeComprobante'].code + ' ' + i['MetodoPago'].code + ' ' + i['FormaPago'].code
    return i['TipoDeComprobante'].code + '       '


def preview_cfdis(cfdis):
    outfile = os.path.join(TEMP_DIRECTORY, "factura.html")
    render.html_write(
        xlm=cfdis,
        target=outfile,
    )
    open_file(
        os.path.abspath(outfile)
    )


def center_location(element) -> tuple[int, int]:
    return tuple((c + s // 3) for c, s in zip(element.current_location(), element.size))



import os
import random
import shutil
import subprocess
import sys
from datetime import datetime, date
from uuid import UUID

from satcfdi.models import Signer, DatePeriod
from satcfdi.accounting.models import EstadoComprobante


def to_date_period(periodo):
    if not periodo:
        return DatePeriod(year=None)

    for f in ('%Y', '%Y-%m', '%Y-%m-%d'):
        try:
            d = datetime.strptime(periodo, f)
            return DatePeriod(
                d.year,
                d.month if '%Y-%m' in f else None,
                d.day if f == '%Y-%m-%d' else None
            )
        except ValueError:
            pass


def to_uuid(s):
    try:
        return UUID(s)
    except ValueError:
        return None


def to_int(s):
    try:
        return int(s)
    except ValueError:
        return None


def random_string():
    chars = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
    return "".join(random.choice(chars) for _ in range(32))


def convert_ans1_date(ans1_date):
    return datetime.strptime(ans1_date.decode('utf-8'), '%Y%m%d%H%M%SZ')


def cert_info(signer: Signer):
    if signer:
        return {
            "NoCertificado": signer.certificate_number,
            # "Tipo": str(signer.type),
            #
            # "organizationName": signer.certificate.get_subject().O,
            # "x500UniqueIdentifier": signer.certificate.get_subject().x500UniqueIdentifier,
            # "serialNumber": signer.certificate.get_subject().serialNumber,
            # "organizationUnitName": signer.certificate.get_subject().OU,
            # "emailAddress": signer.certificate.get_subject().emailAddress,
            #
            "Expira": convert_ans1_date(signer.certificate.get_notAfter()),
            "Creado": convert_ans1_date(signer.certificate.get_notBefore()),
        }


def find_best_match(cases, dp: DatePeriod | date) -> (date, object):
    fk, fv = (None, None)
    for k, v in cases.items():
        k = datetime.strptime(k, '%Y-%m').date()
        if k <= dp:
            if fk is None or k > fk:
                fk, fv = k, v
    return fk, fv


def clear_directory(directory):
    shutil.rmtree(directory, ignore_errors=True)
    os.makedirs(directory, exist_ok=True)


# calculate the number of months between two dates
def months_between(d1, d2):
    return (d1.year - d2.year) * 12 + d1.month - d2.month


def add_month(dp: DatePeriod, months):
    year, month = divmod(dp.year * 12 + dp.month + months - 1, 12)
    month += 1
    return DatePeriod(year, month)


def load_certificate(data):
    if 'data' in data:
        return Signer.load_pkcs12(
            **data,
        )
    else:
        return Signer.load(
            **data,
        )


def first_duplicate(seq):
    seen = set()
    for x in seq:
        if x in seen:
            return x
        seen.add(x)
    return None


def estado_to_estatus(estatus):
    if estatus == 'Vigente':
        return EstadoComprobante.VIGENTE.value
    elif estatus == 'Cancelado':
        return EstadoComprobante.CANCELADO.value
    raise ValueError(f"Unknown status: {estatus}")


def open_file(filename):
    match OS.get_os():
        case OS.WINDOWS:
            os.startfile(filename)
        case OS.LINUX:
            subprocess.call(['xdg-open', filename])
        case OS.MACOS:
            subprocess.call(['open', filename])
        case _:
            raise NotImplementedError(f"Unknown OS: {OS.get_os()}")


# Enum for operative Systems
class OS:
    WINDOWS = 'Windows'
    LINUX = 'Linux'
    MACOS = 'MacOS'
    UNKNOWN = 'Unknown'

    @staticmethod
    def get_os():
        if sys.platform.startswith('win'):
            return OS.WINDOWS
        elif sys.platform.startswith('linux'):
            return OS.LINUX
        elif sys.platform.startswith('darwin'):
            return OS.MACOS
        else:
            return OS.UNKNOWN



import collections.abc
import decimal
import json
import logging
import os
import re
from datetime import date
from decimal import Decimal

import jsonschema as jsonschema
import yaml
from satcfdi.models import Code, DatePeriod
from satcfdi.pacs import sat
from satcfdi.transform.helpers import Xint
from yaml import MappingNode, SafeLoader
from yaml.constructor import ConstructorError

from . import SOURCE_DIRECTORY
from .log_tools import NoAliasDumper
from .utils import find_best_match, first_duplicate

logger = logging.getLogger(__name__)
sat_manager = sat.SAT()


class DuplicateKeySafeLoader(SafeLoader):
    def construct_mapping(self, node, deep=False):
        if isinstance(node, MappingNode):
            self.flatten_mapping(node)

        if not isinstance(node, MappingNode):
            raise ConstructorError(None, None,
                                   "expected a mapping node, but found %s" % node.id,
                                   node.start_mark)
        mapping = {}
        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=deep)
            if not isinstance(key, collections.abc.Hashable):
                raise ConstructorError("while constructing a mapping", node.start_mark,
                                       "found unhashable key", key_node.start_mark)
            value = self.construct_object(value_node, deep=deep)
            if key in mapping:
                raise ConstructorError("while constructing a mapping", node.start_mark,
                                       "found duplicate key (%s)" % key, key_node.start_mark)
            mapping[key] = value
        return mapping


decimal_regex = re.compile(r'[-+]?[0-9_]*\.[0-9_]*', re.X)
for ch in '-+0123456789.':
    DuplicateKeySafeLoader.yaml_implicit_resolvers[ch].insert(
        0,
        (
            '!decimal',
            decimal_regex
        )
    )


def load_validator(schema_file):
    with open(os.path.join(SOURCE_DIRECTORY, 'schemas', schema_file), "r", encoding="utf-8") as fs:
        schema = yaml.load(fs, SafeLoader)
        validator = jsonschema.validators.validator_for(schema)
        validator.check_schema(schema)
        return validator(schema)


client_validator = load_validator("cliente.yaml")
factura_validator = load_validator("factura.yaml")
product_validator = load_validator("producto.yaml")


class LocalData(dict):
    file_source = None

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)

    def __init__(self):
        super().__init__(self._raw())

    def _raw(self):
        try:
            with open(self.file_source, "r", encoding="utf-8") as fs:
                return yaml.load(fs, DuplicateKeySafeLoader)
        except FileNotFoundError:
            os.makedirs(os.path.dirname(self.file_source), exist_ok=True)
            with open(self.file_source, "w", encoding="utf-8") as fs:
                fs.write("{}")
            return {}

    def save(self):
        with open(self.file_source, "w", encoding="utf-8") as fs:
            yaml.dump(self, fs, Dumper=NoAliasDumper, allow_unicode=True, width=1280, sort_keys=False)


class InitManager(LocalData):
    file_source = "init.yaml"

    def __init__(self):
        try:
            super().__init__()
        except FileNotFoundError:
            pass


class ConfigManager(LocalData):
    file_source = "config.yaml"


class ClientsManager(LocalData):
    file_source = "clientes.yaml"

    def __init__(self):
        super().__init__()
        for k, v in self.items():
            if error := jsonschema.exceptions.best_match(client_validator.iter_errors(v)):
                raise error
            self[k]["Rfc"] = k
            self[k]["RegimenFiscal"] = self[k]["RegimenFiscal"]


class ProductosManager(LocalData):
    file_source = "productos.yaml"

    def __init__(self):
        super().__init__()
        if "Constants" in self:
            del self["Constants"]
        for k, v in self.items():
            # if k == 'Constants':
            #     continue
            if error := jsonschema.exceptions.best_match(product_validator.iter_errors(v)):
                raise error


# function to deep merge two dictionaries
def deep_complement(destination, source):
    for key, value in source.items():
        if isinstance(value, dict):
            # get node or create one
            node = destination.setdefault(key, {})
            deep_complement(node, value)
        else:
            destination.setdefault(key, value)


class FacturasManager(LocalData):
    file_source = "facturas.yaml"

    def __init__(self, dp: DatePeriod | date | None):
        def loading_function(loader, node):
            cases = loader.construct_mapping(node, deep=True)
            if dp is None:
                return cases
            return find_best_match(
                cases,
                dp
            )[1]

        DuplicateKeySafeLoader.add_constructor("!case", loading_function)
        super().__init__()
        if dup := first_duplicate(json.dumps(x, sort_keys=True, default=str) for x in self["Facturas"]):
            raise Exception("Factura Duplicada: {}".format(dup))

        pm = ProductosManager()

        for v in self["Facturas"]:
            for c in v["Conceptos"]:
                if p := c.get("_producto"):
                    try:
                        deep_complement(c, pm[p]["Concepto"])
                    except KeyError:
                        raise Exception("Producto no encontrado: {}".format(c["_producto"]))

            if dp:
                if error := jsonschema.exceptions.best_match(factura_validator.iter_errors(v)):
                    raise error


def decimal_constructor(loader, node):
    value = loader.construct_scalar(node)
    try:
        return Decimal(value)
    except decimal.InvalidOperation:
        raise ConstructorError(None, None,
                               "expected a decimal value, but found %s" % value,
                               node.start_mark)


DuplicateKeySafeLoader.add_constructor("!decimal", decimal_constructor)
DuplicateKeySafeLoader.add_constructor("!read", lambda loader, node: open(loader.construct_scalar(node), 'rb').read())


def represent_decimal(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', str(data))


def represent_str(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', str(data))


yaml.SafeDumper.add_multi_representer(dict, lambda dumper, data: dumper.represent_dict(data))
yaml.SafeDumper.add_representer(Decimal, represent_decimal)
yaml.SafeDumper.add_representer(Code, represent_str)
yaml.SafeDumper.add_representer(Xint, represent_str)



from datetime import date, datetime
from decimal import Decimal
from unittest import mock

import pytest
from satcfdi.models import Signer, DatePeriod

from satdigitalinvoice.__version__ import __package__
from satdigitalinvoice.exceptions import ConsoleErrors
from satdigitalinvoice.file_data_managers import ClientsManager, FacturasManager
from satdigitalinvoice.gui_functions import generate_ingresos, periodicidad_desc
from satdigitalinvoice.utils import find_best_match
from tests.utils import verify_result, XElementPrettyPrinter

csd_signer = Signer.load(
    certificate=open('csd/cacx7605101p8.cer', 'rb').read(),
    key=open('csd/cacx7605101p8.key', 'rb').read(),
    password=open('csd/cacx7605101p8.txt', 'rb').read(),
)
ym_date = DatePeriod(year=2023, month=4)
clients = ClientsManager()
module = __package__


def test_generar_ingresos():
    facturas = FacturasManager(ym_date)["Facturas"]

    with mock.patch(f'satcfdi.create.cfd.cfdi40.datetime') as m:
        m.now = mock.Mock(return_value=datetime(2022, 1, 1))

        facturas = generate_ingresos(
            clients=clients,
            facturas=facturas,
            dp=date(year=2023, month=4, day=1),
        )

        assert len(facturas) == 3

        pp = XElementPrettyPrinter()
        verify = verify_result(data=pp.pformat(facturas), filename=f"facturas.pretty.py")
        assert verify


def test_generar_ingresos_error():
    facturas = FacturasManager(ym_date)["FacturasIncorrectas"]
    assert len(facturas) == 1

    with pytest.raises(ConsoleErrors) as e:
        generate_ingresos(
            clients=clients,
            facturas=facturas,
            dp=date(year=2023, month=4, day=1),
        )

    assert e.value.errors == ["1 ABMG891115PD7: Total '41000.16' is invalid, expected '37019.16'"]


def test_generar_ingresos_error2():
    facturas = FacturasManager(ym_date)["FacturasIncorrectas2"]
    assert len(facturas) == 1

    with pytest.raises(ConsoleErrors) as e:
        generate_ingresos(
            clients=ClientsManager(),
            facturas=facturas,
            dp=date(year=2023, month=4, day=1),
        )

    assert e.value.errors == ['1 XXXNOEXISTO: Receptor not found']


def test_periodo_desc():
    assert periodicidad_desc(DatePeriod(year=2021, month=1), 'Mensual.1', offset=0) == 'MES DE ENERO DEL 2021'
    assert periodicidad_desc(DatePeriod(year=2021, month=1), 'Bimestral.2', offset=0) is None
    assert periodicidad_desc(DatePeriod(year=2021, month=1), 'Trimestral.3', offset=0) is None
    assert periodicidad_desc(DatePeriod(year=2021, month=1), 'Cuatrimestral.4', offset=0) is None
    assert periodicidad_desc(DatePeriod(year=2021, month=1), 'Semestral.5', offset=0) is None
    assert periodicidad_desc(DatePeriod(year=2021, month=1), 'Anual.6', offset=0) is None

    assert periodicidad_desc(DatePeriod(year=2021, month=12), 'Mensual.7', offset=0) == 'MES DE DICIEMBRE DEL 2021'
    assert periodicidad_desc(DatePeriod(year=2021, month=12), 'Bimestral.8', offset=0) == 'MES DE DICIEMBRE DEL 2021 AL MES DE ENERO DEL 2022'
    assert periodicidad_desc(DatePeriod(year=2021, month=12), 'Trimestral.9', offset=0) == 'MES DE DICIEMBRE DEL 2021 AL MES DE FEBRERO DEL 2022'
    assert periodicidad_desc(DatePeriod(year=2021, month=12), 'Cuatrimestral.8', offset=0) == 'MES DE DICIEMBRE DEL 2021 AL MES DE MARZO DEL 2022'
    assert periodicidad_desc(DatePeriod(year=2021, month=12), 'Semestral.6', offset=0) == 'MES DE DICIEMBRE DEL 2021 AL MES DE MAYO DEL 2022'
    assert periodicidad_desc(DatePeriod(year=2021, month=12), 'Anual.12', offset=0) == 'MES DE DICIEMBRE DEL 2021 AL MES DE NOVIEMBRE DEL 2022'

    assert periodicidad_desc(DatePeriod(year=2021, month=1), 'Mensual.1', offset=1) == 'MES DE FEBRERO DEL 2021'
    assert periodicidad_desc(DatePeriod(year=2021, month=1), 'Bimestral.2', offset=1) is None
    assert periodicidad_desc(DatePeriod(year=2021, month=1), 'Trimestral.3', offset=1) is None
    assert periodicidad_desc(DatePeriod(year=2021, month=1), 'Cuatrimestral.4', offset=1) is None
    assert periodicidad_desc(DatePeriod(year=2021, month=1), 'Semestral.5', offset=1) is None
    assert periodicidad_desc(DatePeriod(year=2021, month=1), 'Anual.6', offset=1) is None

    assert periodicidad_desc(DatePeriod(year=2021, month=12), 'Mensual.7', offset=1) == 'MES DE ENERO DEL 2022'
    assert periodicidad_desc(DatePeriod(year=2021, month=12), 'Bimestral.8', offset=1) == 'MES DE ENERO DEL 2022 AL MES DE FEBRERO DEL 2022'
    assert periodicidad_desc(DatePeriod(year=2021, month=12), 'Trimestral.9', offset=1) == 'MES DE ENERO DEL 2022 AL MES DE MARZO DEL 2022'
    assert periodicidad_desc(DatePeriod(year=2021, month=12), 'Cuatrimestral.8', offset=1) == 'MES DE ENERO DEL 2022 AL MES DE ABRIL DEL 2022'
    assert periodicidad_desc(DatePeriod(year=2021, month=12), 'Semestral.6', offset=1) == 'MES DE ENERO DEL 2022 AL MES DE JUNIO DEL 2022'
    assert periodicidad_desc(DatePeriod(year=2021, month=12), 'Anual.12', offset=1) == 'MES DE ENERO DEL 2022 AL MES DE DICIEMBRE DEL 2022'


def test_find_best_match():
    casesA = {
        '2022-12': Decimal('20.00'),
        '2023-12': Decimal('30.00'),
        '2024-12': Decimal('40.00'),
        '2025-12': None,
    }
    casesB = {
        '2025-12': None,
        '2024-12': Decimal('40.00'),
        '2023-12': Decimal('30.00'),
        '2022-12': Decimal('20.00'),
    }

    for cases in (casesA, casesB):
        assert find_best_match(cases, DatePeriod(2022, 4, 5))[1] is None
        assert find_best_match(cases, DatePeriod(2023, 4, 5))[1] == Decimal('20.00')
        assert find_best_match(cases, DatePeriod(2024, 4, 5))[1] == Decimal('30.00')
        assert find_best_match(cases, DatePeriod(2025, 4, 5))[1] == Decimal('40.00')
        assert find_best_match(cases, DatePeriod(2026, 4, 5))[1] is None
